# HiveDAV

A curlable free/busy scheduler with CalDAV integration.

## Running
```bash
go run .
```

## Building
```bash
make
```

* The [sqlite3 driver](https://github.com/mattn/go-sqlite3) requires gcc and `CGO_ENABLED=1`

## Linting
```bash
make fmt
```

## Testing
Run tests:
```bash
make test
```

Generate test coverage report:
```bash
make test/cover
```

## Configuration
The application is configurable through environment variables or the
`./app.env` file:

```bash
# Optional
#HIVEDAV_LISTEN_ADDRESS=127.0.0.1
#HIVEDAV_LISTEN_PORT=3737
#HIVEDAV_CALENDAR=0

# Required
HIVEDAV_CALDAV_URI=
HIVEDAV_CALDAV_USER=
HIVEDAV_CALDAV_PASSWORD=
```

There is an example config file provided in `./app.env.example`.

The `HIVEDAV_CALENDAR` is a number that points to the index of the calendar
collection in the ["calendar home
set"](https://www.rfc-editor.org/rfc/rfc4791.html#section-6.2.1) which should
be used to read availability data.

TODO: specify calendar using the "Name" propery

## Storage
The application stores the event start/end times in an sqlite database in the
directory where the process is running.

The schema of the `availability` table is very simple and straight forward:

```bash
sqlite3 app.db ".schema availability"
CREATE TABLE availability (
                        id INTEGER NOT NULL PRIMARY KEY,
                        start DATETIME NOT NULL,
                        end DATETIME NOT NULL,
                        tz TEXT NOT NULL);
```

To show the current availability time slots:
```bash
sqlite3 app.db "select * from availability;" | less
sqlite3 app.db "select * from availability where start between '2023-01-01' and '2023-02-01';" | less
```

There exists another table `availability_1` which is used to store the new
availability time slots during an availability update ("CalDAV request").
During this CalDAV request, the table `availability_1` is used to temporarily
store the new availability. After the update is done, the temporary table is
set as the "current" availability. The application only reads availability from
the `availability` table, not from the temporary `availability_1` table.

## CalDAV Requests
The CalDAV requests in the `caldav` module are used to query:
* the user principal
* the calendar home set
* the calendar from the home set (`HIVEDAV_CALENDAR`) used as source of
  availability data

The CalDAV requests are mad according to the examples in
https://sabre.io/dav/building-a-caldav-client.

The requests are kept very simple intentionally with libraries included in Go:
* The requests only use the `net/http` library to do the HTTP requests.
* The response is parsed using the `encoding/xml` library

No external library was used perform the requests, because these libraries
typically assume specific behavior for the remote CalDAV server. Unfortunately,
the servers vary in their behavior quite a lot. For instance, certain servers
don't support the [`calendar-query
REPORT`](https://www.rfc-editor.org/rfc/rfc4791.html#section-7.8). To make this
application work with as many remote servers as possible, all events from the
remote calendar are fetched during an update and the availability data is
stored in a local database for further processing.

## Availability Query

```
//  e ?   ?
//  - |   |   false

//  s e   ?
//  - +   |   false

//  s ? e ?
//  - | - |   true (2)

//  s ?   e
//  - |   +   true (2)

//  s ?   ? e
//  - |   | - true (3)

//    s e ?
//    + - |   true (1)

//    s   e
//    +   +   true (2)

//    s   ? e
//    +   | - true (1)

//    ? s e
//    | - +   true (1)

//    ? s ? e
//    | - | - true (1)

//    ?   s e
//    |   + - false
```

TODO:
- Sqlite date comparison always inclusive?
- Picture for the query

## Development & Contributions
* All source code is available in this repository. Contributions are always
  welcome!
* A [Kanban
  board](https://board.in0rdr.ch/public/board/01171363fae4b8b81af1579708f7cda6d389062f05b75db860621eaed32b)
documents plans, todos and decisions for further development of the project.

## Community and Support
For question and support visit
[#hivedav/libera](https://web.libera.chat/gamja/#hivedav) on IRC.
