package caldav

// Functions to read VEVENTs from the CalDav server:
// https://sabre.io/dav/building-a-caldav-client

import (
	"bytes"
	"encoding/xml"
	"fmt"
	"net/http"
)

type XMLUserPrincipalResponse struct {
	XMLName       xml.Name `xml:multistatus`
	UserPrincipal string   `xml:"response>propstat>prop>current-user-principal>href"`
}
type XMLHomeSetResponse struct {
	XMLName xml.Name `xml:multistatus`
	HomeSet string   `xml:"response>propstat>prop>calendar-home-set>href"`
}
type XMLHomeSetCalendarResponse struct {
	XMLName  xml.Name   `xml:multistatus`
	Response []Response `xml:"response"`
}
type Response struct {
	XMLName xml.Name `xml:response`
	Href    string   `xml:"href"`
}
type XMLCalendarResponse struct {
	XMLName xml.Name  `xml:multistatus`
	CalData []CalData `xml:"response"`
}
type CalData struct {
	XMLName xml.Name `xml:response`
	Href    string   `xml:"href"`
	Data    string   `xml:"propstat>prop>calendar-data"`
}

func UserPrincipal(caldavUri string, user string, pass string) (string, error) {
	payload := []byte(`<d:propfind xmlns:d="DAV:">
		      <d:prop>
		         <d:current-user-principal />
		      </d:prop>
		    </d:propfind>`)
	reader := bytes.NewReader(payload)
	req, err := http.NewRequest("PROPFIND", caldavUri, reader)
	if err != nil {
		return "", err
	}
	req.Header.Set("Depth", "0")
	req.SetBasicAuth(user, pass)

	res, err := http.DefaultClient.Do(req)
	if err != nil {
		return "", err
	}

	if res.StatusCode != 207 {
		return "", err
	}

	decoder := xml.NewDecoder(res.Body)
	var xmlResponse XMLUserPrincipalResponse
	err = decoder.Decode(&xmlResponse)
	if err != nil {
		return "", err
	}

	return xmlResponse.UserPrincipal, nil
}

func CalendarHome(caldavUri string, user string, pass string, userPrincipal string, calendar int) (string, error) {
	payload := []byte(`<d:propfind xmlns:d="DAV:" xmlns:c="urn:ietf:params:xml:ns:caldav">
			     <d:prop>
				<c:calendar-home-set />
			     </d:prop>
			   </d:propfind>`)
	reader := bytes.NewReader(payload)
	principalUri := fmt.Sprintf("%s%s", caldavUri, userPrincipal)

	req, err := http.NewRequest("PROPFIND", principalUri, reader)
	if err != nil {
		return "", err
	}
	req.Header.Set("Depth", "0")
	req.SetBasicAuth(user, pass)

	res, err := http.DefaultClient.Do(req)
	if err != nil {
		return "", err
	}

	if res.StatusCode != 207 {
		return "", err
	}

	decoder := xml.NewDecoder(res.Body)
	var xmlResponse XMLHomeSetResponse
	err = decoder.Decode(&xmlResponse)
	if err != nil {
		return "", err
	}

	return xmlResponse.HomeSet, nil
}

func CalendarFromHomeSet(caldavUri string, user string, pass string, homeSetUri string, calendar int) (calendarUri string, err error) {
	payload := []byte(`<d:propfind xmlns:d="DAV:" xmlns:cs="http://calendarserver.org/ns/" xmlns:c="urn:ietf:params:xml:ns:caldav">
			     <d:prop>
			        <d:resourcetype />
			        <d:displayname />
			        <cs:getctag />
			        <c:supported-calendar-component-set />
			     </d:prop>
			   </d:propfind>`)
	reader := bytes.NewReader(payload)
	homeSetUrl := fmt.Sprintf("%s%s", caldavUri, homeSetUri)

	req, err := http.NewRequest("PROPFIND", homeSetUrl, reader)
	if err != nil {
		return "", err
	}
	req.Header.Set("Depth", "1")
	req.SetBasicAuth(user, pass)

	res, err := http.DefaultClient.Do(req)
	if err != nil {
		return "", err
	}

	if res.StatusCode != 207 {
		return "", err
	}

	decoder := xml.NewDecoder(res.Body)
	var xmlResponse XMLHomeSetCalendarResponse
	err = decoder.Decode(&xmlResponse)
	if err != nil {
		return "", err
	}

	defer func() {
		if r := recover(); r != nil {
			err = fmt.Errorf("Calendar '%d' not found in homeset: %v", calendar, r)
		}
	}()

	calendarUri = xmlResponse.Response[calendar].Href
	return calendarUri, nil
}

func GetAvailability(caldavUri string, user string, pass string, calendarUri string) (data []CalData, err error) {
	payload := []byte(`<c:calendar-query xmlns:d="DAV:" xmlns:c="urn:ietf:params:xml:ns:caldav">
			       <d:prop>
			           <d:getetag />
			           <c:calendar-data />
			       </d:prop>
			       <c:filter>
			           <c:comp-filter name="VCALENDAR">
			               <c:comp-filter name="VEVENT" />
			           </c:comp-filter>
			       </c:filter>
			   </c:calendar-query>`)
	reader := bytes.NewReader(payload)
	calendarUrl := fmt.Sprintf("%s%s", caldavUri, calendarUri)

	req, err := http.NewRequest("REPORT", calendarUrl, reader)
	if err != nil {
		return data, err
	}
	req.Header.Set("Depth", "1")
	req.SetBasicAuth(user, pass)

	res, err := http.DefaultClient.Do(req)
	if err != nil {
		return data, err
	}

	if res.StatusCode != 207 {
		return data, err
	}

	decoder := xml.NewDecoder(res.Body)
	var xmlResponse XMLCalendarResponse
	err = decoder.Decode(&xmlResponse)
	if err != nil {
		return data, err
	}

	data = xmlResponse.CalData
	return data, nil
}
