package main

import (
	"fmt"
	"log"
	"net/http"
	"hivedav/config"
	"hivedav/caldav"
	"github.com/julienschmidt/httprouter"
)

func main() {
	// If exists, load config file from pwd
	conf := &config.Config{}
	conf, err := conf.LoadConfig(".")
	if err != nil {
		log.Fatal("Error loading config: ", err)
	}

	var server *Server
	server, err = server.NewServer(conf)
	if err != nil {
		log.Fatal("Error creating server with config: ", err)
	}

	log.Println("Starting hivedav v0.1 🍯")

	log.Println("Reading availability from:", conf.CaldavUri)
	log.Println("Reading as user:", conf.CaldavUser)

	// Find current user principal
	userPrincipal, err := caldav.UserPrincipal(conf.CaldavUri, conf.CaldavUser, conf.CaldavPassword)
	if err != nil {
		log.Fatal("Error reading current user principal: ", err)
	}
	log.Println("User principal:", userPrincipal)

	// Find home set of current user principal
	homeSetUri, err := caldav.CalendarHome(conf.CaldavUri, conf.CaldavUser, conf.CaldavPassword, userPrincipal, conf.Calendar)
	if err != nil {
		log.Fatal("Error reading home set URI: ", err)
	}
	log.Println("Calendar homeset:", homeSetUri)

	// Find calendar URI for the calendar specified in the config
	calendarUri, err := caldav.CalendarFromHomeSet(conf.CaldavUri, conf.CaldavUser, conf.CaldavPassword, homeSetUri, conf.Calendar)
	if err != nil {
		log.Fatal("Error reading calendar URI: ", err)
	}
	log.Printf("Using calendar %d in homeset with URI '%s'\n", conf.Calendar, calendarUri)
	log.Println("⏳ Fetching availability from CalDAV server, please wait..")

	// Get availability data for the calendar specified in the config
	calData, err := caldav.GetAvailability(conf.CaldavUri, conf.CaldavUser, conf.CaldavPassword, calendarUri)
	if err != nil {
		log.Fatal("Error reading calendar URI: ", err)
	}

	err = server.updateAvailability(calData)
	if err != nil {
		log.Fatalf("Error fetching availability: %v", err)
	}

	log.Printf("Ready, listening on %s:%d\n", conf.ListenAddress, conf.ListenPort)

	router := httprouter.New()
	router.GET("/", server.Index)
	router.GET("/week/:week", server.Week)

	log.Fatal(http.ListenAndServe(fmt.Sprintf("%s:%d", conf.ListenAddress, conf.ListenPort), router))
}
