package main

import (
	"regexp"
	"time"
	"strconv"
	"github.com/teambition/rrule-go"
	"github.com/emersion/go-ical"
	"hivedav/config"
	"hivedav/caldav"
	"fmt"
	"github.com/pquerna/termchalk/ansistyle"
	"github.com/pquerna/termchalk/prettytable"
	_ "github.com/mattn/go-sqlite3"
	"database/sql"
	"io"
	"log"
	"net/http"
	"github.com/julienschmidt/httprouter"
	"strings"
)

type Server struct {
	config *config.Config
	db *sql.DB
}

var drop = `DROP TABLE IF EXISTS availability;
	    DROP TABLE IF EXISTS availability_1;`
var create_availability = `CREATE TABLE IF NOT EXISTS availability (
			   id INTEGER NOT NULL PRIMARY KEY,
			   start DATETIME NOT NULL,
			   end DATETIME NOT NULL,
			   recurring BOOL NOT NULL);`
var create_availability_1 = `CREATE TABLE IF NOT EXISTS availability_1 (
			     id INTEGER NOT NULL PRIMARY KEY,
			     start DATETIME NOT NULL,
			     end DATETIME NOT NULL,
			     recurring BOOL NOT NULL);`

// BETWEEN is inclusive, we tweak start/time by 1 min to make the edge cases
// 08:01, 09:59, etc. behave correctly
var available = `SELECT id FROM availability WHERE
		 (start BETWEEN ? AND ?) OR
		 (end BETWEEN ? AND ?) OR
		 (start <= ? AND end >= ?);`

func (s *Server) NewServer(c *config.Config) (*Server, error) {
	s = new(Server)
	s.config = c

	// prepare the sqlite database
	db, err := sql.Open("sqlite3", "app.db")
	if err != nil {
		log.Fatal("Error opening sqlite3 database 'app.db'", err)
	}
	s.db = db

	if _, err := db.Exec(drop); err != nil {
		return nil, err
	}
	if _, err := db.Exec(create_availability); err != nil {
		return nil, err
	}
	if _, err := db.Exec(create_availability_1); err != nil {
		return nil, err
	}

	return s, nil
}

// https://xferion.com/golang-reverse-isoweek-get-the-date-of-the-first-day-of-iso-week
func dayOfISOWeek(weekday int, week int, year int) time.Time {
	timeNow := time.Now()
	yearNow, weekNow:= timeNow.ISOWeek()

	// iterate back to the weekday
	for timeNow.Weekday() != time.Weekday(weekday) {
		// years, months, days
		timeNow = timeNow.AddDate(0, 0, -1)
		yearNow, weekNow = timeNow.ISOWeek()
	}

	// if we iterated back into the old year, iterate forward in weekly
	// steps (7) until we reach the weekday of the first week in yearNow
	for yearNow < year {
		timeNow = timeNow.AddDate(0, 0, 7)
		yearNow, weekNow = timeNow.ISOWeek()
	}

	// iterate forward to the first day of the given ISO week
	for weekNow < week {
		timeNow = timeNow.AddDate(0, 0, 7)
		yearNow, weekNow = timeNow.ISOWeek()
	}

	return timeNow
}

func (s *Server) available(start string, end string) (bool, error) {
	var id int

	stmt, err := s.db.Prepare(available)

	if err != nil {
		log.Println("Error during db statement Prepare()")
		return false, err
	}

	err = stmt.QueryRow(start, end, start, end, start, end).Scan(&id)
	if err != nil {
		if err == sql.ErrNoRows {
			// We are available if we cannot find any rows
			return true, nil
		}
		return false, err
	}
	return false, nil
}

func (s *Server) Week(w http.ResponseWriter, req *http.Request, ps httprouter.Params) {
	// TODO: make display timezone configurable, display in local time
	localLocation, err := time.LoadLocation("Local")
	if err != nil {
		return
	}

	userAgent := req.Header.Get("user-agent")

	timeNow := time.Now()
	weekday := timeNow.Weekday()
	year, week := timeNow.ISOWeek()

	// overwrite with requested week
	week, err = strconv.Atoi(ps.ByName("week"))
	if err != nil {
		log.Printf("Cannot serve requested week '%s'")
		w.WriteHeader(http.StatusBadRequest)
		return
	}

	log.Printf("Current weekday is '%d'\n", weekday)
	log.Printf("Serving week '%v' of year '%v'\n", week, year)

	monday := dayOfISOWeek(1, week, year)
	log.Printf("Monday is '%v'\n", monday)
	sqlDateFmt := "2006-01-02 15:04"

	if strings.Contains(userAgent, "curl") {
		w.Header().Set("Content-Type", "text/plain")
		io.WriteString(w, ansistyle.Bold.Open+"Welcome to the "+ansistyle.Bold.Close)
		io.WriteString(w, ansistyle.BgRed.Open+"hive!"+ansistyle.BgBlue.Close+"\n")
		io.WriteString(w, fmt.Sprintf("Serving week %d of year %d\n", week, year))

		// TODO: Make heading date format configurable, start of week, etc..
		pt := prettytable.New([]string{
			"Time      ",
			fmt.Sprintf("Mon %s", monday.Format("02.01.")),
			fmt.Sprintf("Tue %s", monday.Add(time.Hour).Format("02.01.")),
			fmt.Sprintf("Wed %s", monday.Add(time.Hour*24*2).Format("02.01.")),
			fmt.Sprintf("Thu %s", monday.Add(time.Hour*24*3).Format("02.01.")),
			fmt.Sprintf("Fri %s", monday.Add(time.Hour*24*4).Format("02.01.")),
		})

		monTime:= dayOfISOWeek(1, week, year)
		// TODO: use timeIt to go through the loops below, remove the static arrays
		timeIt := time.Date(monTime.Year(), monTime.Month(), monTime.Day(), 0, 0, 0, 0, localLocation)
		//dayEnd := time.Date(monTime.Year(), monTime.Month(), monTime.Day(), 17, 0, 0, 0, localLocation)
		//weekEnd := dayEnd.Add(time.Hour * 24*5)

		// Working hours - Eight to Five
		//for h := 8; h < 17; h++ {
		for _, h := range []int{8,9,10,11,12,13,14,15,16} {
			var availability [5]string
			// Working days - Monday through Friday
			//for d := 0; d < 5; d++ {
			for _, d := range []int{1,2,3,4,5} {
				// add/remove 1 minute to make the sql time query in available() below behave correctly
				// convert to read/compare in UTC from db
				start := timeIt.Add(time.Hour * time.Duration((d-1)*24 + h)).Add(1 * time.Minute).UTC().Format(sqlDateFmt)
				end := timeIt.Add(time.Hour * time.Duration((d-1)*24 + h+1)).Add(-1 * time.Minute).UTC().Format(sqlDateFmt)
				avi, err := s.available(start, end)
				log.Printf("Getting availability on day '%v' - '%v'\n", start, end)
				if err != nil {
					log.Printf("Error getting availability on day '%d' hour '%d'\n", d, h)
					return
				}
				if avi {
					availability[d-1] = ""
				} else {
					availability[d-1] = "X"
				}
				//timeIt = timeIt.Add(time.Hour * 24)
			}
			pt.AddRow(fmt.Sprintf("%02d:00 - %02d:00", h, h+1),
				availability[0],
				availability[1],
				availability[2],
				availability[3],
				availability[4])
			//timeIt = timeIt.Add(time.Hour)
		}

		io.WriteString(w, fmt.Sprint(pt))
	} else {
		w.Header().Set("Content-Type", "text/html")
		io.WriteString(w, "<h1>Welcome to the hive!</h1>")
		io.WriteString(w, "<p>HTML Calendar here</p>")
	}
}

func (s *Server) Index(w http.ResponseWriter, req *http.Request, _ httprouter.Params) {
	timeNow := time.Now()
	_, currentWeek:= timeNow.ISOWeek()
	ps := httprouter.Params{
		httprouter.Param{"week", strconv.Itoa(currentWeek)},
	}
	s.Week(w, req, ps)
}

func (s *Server) updateAvailability(calData []caldav.CalData) error {
	// TODO: Make max for recurring events without end time configurable
	rrLimit := time.Date(2024, 12, 31, 0, 0, 0, 0, time.Local)
	//localNow := time.Now()

	// Walk through events and store start/end time in the database
	for _, event := range calData {
		log.Printf("%v\n", event.Data)
		dec := ical.NewDecoder(strings.NewReader(event.Data))
		cal, err := dec.Decode()
		var (
			//tzProp *ical.Prop
			// offset which is in use prior to this time zone observance
			//tzOffsetTo string
			//tzOffsetFrom string
			// offset which is in use in this time zone observance
			tzOffsetTo string
			// TODO:TZOFFSETFROM is the local time offset from GMT
			// when daylight saving time is in operation,
			// TZOFFSETTO is the local time offset from GMT when
			// standard time is in operation
			// https://stackoverflow.com/questions/3872178/what-does-tzoffsetfrom-and-tzoffsetto-mean
		)

		if err != nil {
			return err
		}

		// check for TIMEZONE component
		for _, c := range cal.Children {
			if c.Name == ical.CompTimezone {
				for _, child := range c.Children {
					//tzOffsetFrom = child.Props.Get(ical.PropTimezoneOffsetFrom).Value;
					tzOffsetTo = child.Props.Get(ical.PropTimezoneOffsetTo).Value;
					log.Println("tzOffsetTo:", tzOffsetTo)
				}
			}
		}

		var tzOffsetDuration time.Duration
		re := regexp.MustCompile(`^([+-][0-9]{2}).*`)
    		matches := re.FindStringSubmatch(tzOffsetTo)
		if len(matches) == 0 {
			//_, offset := localNow.Zone()
			//tzOffsetDuration = time.Second * time.Duration(offset)
			//tzOffsetDuration = time.Second * time.Duration(-1 * offset)
			//log.Printf("Cannot find time zone offset, using local offset: %v\n", tzOffsetDuration)
			tzOffsetDuration = time.Duration(0)
			log.Println("Cannot find time zone offset, using 0 offset.")
		} else {
			tzOffsetDurationInt, _ := strconv.Atoi(matches[1])
			tzOffsetDuration = time.Hour * time.Duration(tzOffsetDurationInt)
    			log.Printf("Found time zone offset: %v\n", tzOffsetDuration)
			// reset to UTC
			tzOffsetDuration = time.Hour * time.Duration(-1 * tzOffsetDurationInt)
		}

		for _, e := range cal.Events() {
			start := e.Props.Get(ical.PropDateTimeStart).Value
			end := e.Props.Get(ical.PropDateTimeEnd).Value

			// Parse all possible time formats
			// Don't use ical DateTime(), has problems with some timezones
			startTime, err := parseTime(start, tzOffsetDuration)
			if err != nil {
				return err
			}
			endTime, err := parseTime(end, tzOffsetDuration)
			if err != nil {
				return err
			}
			duration := endTime.Sub(startTime)

			// Parse the recurrence rules
			// https://github.com/emersion/go-ical/blob/master/ical.go
			// https://github.com/emersion/go-ical/blob/master/components.go
			roption, err := e.Props.RecurrenceRule()
			if err != nil {
				return err
			}

			// set new availability in temporary table availability_1
			if _, err = s.db.Exec("INSERT INTO availability_1 VALUES(NULL, ?, ?, false);", startTime, endTime); err != nil {
				return err
			}

			// Save recurring events
			if roption != nil {
				r, err := rrule.NewRRule(*roption)
				if err != nil {
					return err
				}

				set := rrule.Set{}
				set.RRule(r)
				set.DTStart(startTime)

				// add exception dates to recurrence set exclusion list
				exceptionDates := e.Props.Get(ical.PropExceptionDates)
				if exceptionDates != nil {
					exDateTime, err := parseTime(exceptionDates.Value, tzOffsetDuration)
					if err != nil {
						return err
					}
					set.ExDate(exDateTime)
				}

				// don't include the first event, handled outside the loop
				rDates := set.Between(startTime, rrLimit, false)
				for _, rrd := range rDates {
					if _, err = s.db.Exec("INSERT INTO availability_1 VALUES(NULL, ?, ?, true);", rrd, rrd.Add(duration)); err != nil {
						return err
					}
				}
			}
		}
	}

	// delete current availability
	if _, err := s.db.Exec("DROP TABLE IF EXISTS availability"); err != nil {
		return err
	}
	// set new availability as current availability
	if _, err := s.db.Exec("ALTER TABLE availability_1 RENAME TO availability"); err != nil {
		return err
	}
	// prepare new temporary table
	if _, err := s.db.Exec(create_availability_1); err != nil {
		return err
	}

	return nil
}

// parse time from local time zone and return UTC time for sqlite db
func parseTime(timeStr string, offset time.Duration) (time.Time, error){
	t, err := time.Parse("20060102T150405Z", timeStr)
	if err != nil {
		t, err = time.Parse("20060102T150405", timeStr)
	}
	if err != nil {
		t, err = time.Parse("20060102", timeStr)
	}
	if err != nil {
		return time.Time{}, err
	}
	return t.Add(offset), nil
}

func (s *Server) showCubicles() {
}

func (s *Server) reserveCubicle() {
}
